/*
 *  sendmail.cpp
 *  Copyright (C) 2010 Katsuhisa Ohfuji <katsuhisa@ohfuji.name>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */
#include <iostream>
#include <string>
#include <vector>

#ifdef _WIN32
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#else
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define SOCKET	int
#define closesocket	close

#endif

using namespace std;

#pragma comment( lib, "ws2_32.lib" ) 

static int tcp_connect(SOCKET &sofd, const char* host, int port)
{
	struct sockaddr_in	sv_addr;
	struct hostent		*shost;

    /* \bPg̍쐬 */
    sofd = socket( PF_INET, SOCK_STREAM, 0);
    if ( sofd == -1 ) {
		return -1;
    }

    /* T[õzXghoAhX𓾂 */
    shost = gethostbyname(host);
    if ( shost == NULL ) {
	    closesocket(sofd);
        return -1;
    }

    /* T[oconnect */
    memset((char *)&sv_addr, '\0', sizeof(sv_addr));
    sv_addr.sin_family      = PF_INET;
    sv_addr.sin_port        = htons(port);
    memcpy((char *)&sv_addr.sin_addr, (char *)(shost->h_addr), shost->h_length);
    if ( connect( sofd, (struct sockaddr *)&sv_addr, sizeof(sv_addr)) < 0 ) {
	    closesocket(sofd);
        return -1;
    }

	return 0;
}

static int tcp_send_text(SOCKET &sofd, const string& sbuf)
{
	if ( sbuf.empty() ) return 0;

	int	len;
	/* NGXg̑M */
	len = send( sofd, sbuf.c_str(), sbuf.length(), 0);
    if ( len < 0 ) {
        return -1;
    }

	return 0;
}

static int tcp_recv_text(SOCKET &sofd, string& rbuf)
{
	int	len;
	/* X|X̎M */
	rbuf.clear();
	rbuf.reserve(204800);
	char	buf[102400];
	while ( (len = recv( sofd, buf, sizeof(buf) - 1, 0)) > 0 ) {
		buf[len] = '\0';
		rbuf += buf;
	}

	return len >= 0 ? 0 : -1;
}

static int tcp_recv_line(SOCKET &sofd, string& rbuf)
{
	char	buf[1];
	int		len = -1;
	rbuf.clear();
	rbuf.reserve(2048);
	while ( (len = recv( sofd, buf, 1, 0)) > 0 ) {
		rbuf.push_back(buf[0]);
		if ( buf[0] == '\n' ) return 0;
	}
	return len >= 0 ? 0 : -1;
}

static int tcp_close(SOCKET &sofd) 
{
	closesocket(sofd);
	return 0;
}

static string base64encord( const char *src)
{
	static char	etbl[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	string dst;

	while ( *src ) {
		unsigned long eval = 0;
		int cnt = 0;
		while ( cnt < 3 && *src ) {
			eval <<= 8;
			eval |= *src++;
			cnt++;
		}
		eval <<= 8 * (3 - cnt);
		dst.push_back( etbl[(eval >> 18) & 0x3f]);
		dst.push_back( etbl[(eval >> 12) & 0x3f]);
		if ( cnt <= 1 ) {
			dst.push_back('=');
		} else {
			dst.push_back(etbl[(eval >>  6) & 0x3f]);
		}
		if ( cnt <= 2 ) {
			dst.push_back('=');
		} else {
			dst.push_back(etbl[eval & 0x3f]);
		}
	}
	return dst;
}

static string base64decord( const char *src )
{
	static char dtbl[] =
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3e\x00\x00\x00\x3f"
	"\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e"
	"\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x00\x00\x00\x00\x00"
	"\x00\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28"
	"\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00";

	string			dst;
	unsigned int	dval = 0;
	unsigned int	oval;
	int	len = 0;
	int scnt = 0;

	while ( *src && *src != '=' ) {
		dval <<= 6;
		dval |= dtbl[*src];
		scnt += 6;
		if ( scnt >= 8 ) {
			oval = dval >> (scnt - 8);
			dval &= (1 << ((scnt - 8) + 1)) - 1;
			dst.push_back(oval);
			len++;
			scnt -= 8;
		}
		src++;
	}
	return dst;
}


static bool smtp_check_response(SOCKET sofd, int expect)
{
	string res;
	if ( tcp_recv_line(sofd, res) < 0 ) return false;
	if ( atoi(res.c_str()) != expect ) return false;
	
	return true;
}

static bool smtp_send_crlf(SOCKET sofd)
{
	string	crlf = "\r\n";
	if ( tcp_send_text( sofd, crlf) != 0 ) return false;
	return true;
}

static bool smtp_send_line(SOCKET sofd, const string &cmd, const string &param)
{
	if ( tcp_send_text( sofd, cmd) != 0 ) return false;
	if ( tcp_send_text( sofd, param) != 0 ) return false;
	return smtp_send_crlf(sofd);
}

static bool smtp_send_command(SOCKET sofd, const string &cmd, const string &param, int expect)
{
	if ( !smtp_send_line( sofd, cmd, param) ) return false;
	return smtp_check_response( sofd, expect);
}

inline bool smtp_send_header_b64encode( SOCKET sofd, const string &cmd, const string &param, const string &charset) {

	string b64param = " =?" + charset + "?B?" + base64encord(param.c_str()) + "?=";
	if ( !smtp_send_line( sofd, cmd, b64param) ) return false;
	return true;
}

inline bool smtp_send_mail_text(SOCKET sofd, const string &to, const string &cc, const string &subject, const string &text, const string &charset ) {

	if ( !to.empty() && !smtp_send_header_b64encode( sofd, "To:", to, charset) ) return false;
	if ( !cc.empty() && !smtp_send_header_b64encode( sofd, "Cc:", cc, charset) ) return false;
	if ( !subject.empty() && !smtp_send_header_b64encode( sofd, "Subject:", subject, charset ) ) return false;
	if ( !smtp_send_line( sofd, "MIME-Version:", " 1.0") ) return false;
	if ( !smtp_send_line( sofd, "Content-Type:", " text/plain; charset=" + charset ) ) return false;
	if ( !smtp_send_line( sofd, "Content-Transfer-Encoding:", " base64") ) return false;
	if ( !smtp_send_crlf( sofd) ) return false;
	if ( tcp_send_text( sofd, base64encord(text.c_str())) != 0 ) return false;
	if ( !smtp_send_crlf( sofd) ) return false;
	if ( send( sofd, ".", 1, 0) != 1 ) return false;
	if ( !smtp_send_crlf( sofd) ) return false;

	return smtp_check_response(sofd, 250);
}

inline bool sendmail( const string &smtpserver, const string &to, const string &cc, const string &bcc, const string &from, const string &subject, const string &text, const string &charset ) {
	SOCKET	sofd;
	bool	result = false;

	if ( tcp_connect( sofd, smtpserver.c_str(), 25) != 0 ) return false;

	char	hostname[256];
	if ( gethostname( hostname, sizeof(hostname)) != 0 ) return false;

	string recept_str = to;
	if ( !cc.empty() ) recept_str = recept_str + "," + cc;
	if ( !bcc.empty() ) recept_str = recept_str + "," + bcc;
	vector<string>	recpt_to;
	size_t			off = 0;
	size_t			fnd = 0;
	while ( (fnd = recept_str.find( ',', off)) != string::npos ) {
		string recept(recept_str.begin() + off, recept_str.begin() + fnd);
		recpt_to.push_back(recept);
		off = fnd + 1;
	}
	string recept( recept_str.begin() + off, recept_str.end());
	recpt_to.push_back(recept);

    if ( !smtp_check_response( sofd, 220) ) goto error_out;
    if ( !smtp_send_command( sofd, "HELO ", hostname, 250) ) goto error_out;
    if ( !smtp_send_command( sofd, "MAIL FROM:", from, 250) ) goto error_out;

	for ( vector< string >::iterator i = recpt_to.begin(); i < recpt_to.end(); i++ ) {
        if ( !smtp_send_command( sofd, "RCPT TO:", (*i).c_str(), 250) ) goto error_out;
	}
    if ( !smtp_send_command( sofd, "DATA", "", 354) ) goto error_out;
    if ( !smtp_send_mail_text( sofd, to, cc, subject, text, charset) ) goto error_out;
    if ( !smtp_send_command( sofd, "QUIT", "", 221) ) goto error_out;
    result = true;
error_out:
	tcp_close(sofd);
	return result;
}


int main(void)
{
#ifdef _WIN32
	WSADATA				wsaData;
    WSAStartup( 0x0101, &wsaData);
	string	charset = "Shift_JIS ";
#else
	string	charset = "utf-8";
#endif

	string	smtpserver("mail");
	string	to("to_address@example.com");
	string	cc("");
	string	bcc("");
	string  from("from_address@example.com");

	string	subject("TuWFNg");
	string	text("[{");

	cout << subject << ":" << text << endl;

	if ( !sendmail( smtpserver, to, cc, bcc, from, subject, text, charset) ) {
		cout << "Error Send Mail.";
		return 1;
	}

#ifdef _WIN32
    WSACleanup();
#endif

	return 0;
}

